Unit FtpSrv;
{$I-}

Interface

uses
    Dos, crt, Objects, SysUtils, Sockets, IP, Network, Util;

Type
    TFTPConnectionFlags = (ftpBinary);
    TFTPConnectionStatus = (stBanner, stWaitingUser, stWaitingPassword,
      stWaitingCommand, stSendingDir, stSendingList, stReceivingFile,
      stSendingFile, stDone);

    TFTPConnection = class
    private
        FFlags: set of TFTPConnectionFlags;
        FLastComm: string[96];
        FCtrlSocket: Integer;      	// Control Connection
        FDataSocket: Integer;      	// Data Connection
        FCloseTimeout: Comp;
        FUserName: AnsiString;
        FRootPath: AnsiString;
        FSearchRec: TSearchRec;
        FCurrentPath: AnsiString;
        FDataPort: TInetSockAddr;   	// Remote port information
        FRemotePeer: TInetSockAddr;		// Remote peer information
        FCurrentFileHandle: Integer;
        FCurrentFileName: AnsiString;
        FStatus: TFTPConnectionStatus;
        procedure ProcessControlLine(Line: AnsiString);
        procedure AdjustPath(var Path: AnsiString);
        function  UnixFileName(Path: AnsiString): AnsiString;
    protected
    	procedure Activity;
        procedure SendBanner;
        procedure SendDirBlock;
        procedure SendListBlock;
        procedure SendFileBlock;
        procedure ReceiveFileBlock;
        procedure CheckControlSocket;
        function  OpenDataConnection: Boolean;
        procedure SendResponse(Response: AnsiString);
        //
        Procedure Proc_Pwd(Params: AnsiString);
        Procedure Proc_Cwd(Params: AnsiString);
        Procedure Proc_Mkd(Params: AnsiString);
        Procedure Proc_Rmd(Params: AnsiString);
        Procedure Proc_Help(Params: AnsiString);
        Procedure Proc_User(Params: AnsiString);
        Procedure Proc_Pass(Params: AnsiString);
        Procedure Proc_Port(Params: AnsiString);
        Procedure Proc_Type(Params: AnsiString);
        Procedure Proc_RnFr(Params: AnsiString);
        Procedure Proc_Rnto(Params: AnsiString);
        Procedure Proc_List(Params: AnsiString);
        Procedure Proc_Nlst(Params: AnsiString);
        Procedure Proc_Stor(Params: AnsiString);
        Procedure Proc_Dele(Params: AnsiString);
        Procedure Proc_Retr(Params: AnsiString);
        Procedure Proc_Site(Params: AnsiString);
        Procedure Proc_Size(Params: AnsiString);
        Procedure Proc_Quit(Params: AnsiString);
        Procedure Proc_Abor(Params: AnsiString);
    public
        constructor Create(Socket: Integer; RemotePeer: TInetSockAddr; RootPath: AnsiString);
        destructor Destroy; override;
        procedure Work;

        property Status: TFTPConnectionStatus Read FStatus;
    end;

    TFTPServer = class
    private
        FSocket: Integer;
        FRootPath: AnsiString;
        FConnections: PCollection;
    protected
        procedure AddConnection(Socket: Integer; Addr: TInetSockAddr);
    public
        constructor Create;
        destructor Destroy; override;
        function Start(Address: TIPAddress; Port: WORD; RootPath: AnsiString): Boolean;
        procedure Work;
    end;

implementation

const
	CLOSE_TIMEOUT = 60000 * 5;

    // 100
    MSG_ALREADYCONNECT	= '125 Already connected. Starting Transfer.';
    MSG_CONNECTING      = '150 Opening data connection.';

    // 200
    MSG_NOOP            = '200 Ok.';
    MSG_SYST            = '215 DOS GO32V2/FreePascal.';
    MSG_SIZE            = '213 %d';
    MSG_PORTOK          = '200 PORT command successful.';
	MSG_MKDOK 			= '200 Directory %s created.';
    MSG_TYPEOK          = '200 Type set to %s.';
    MSG_BANNER          = '220-FreePascal/DOS FTP Server ready.' + CRLF +
                          '220-Original version by Xon Electronics SRL (www.xonelectronics.it).' + CRLF +
                          '220 Version 2.0, September 2001 by Diego De Marco (www.dpsautomation.com).';
    MSG_BYE             = '221 Goodbye.';
	MSG_ENDDIR    		= '226 Transfer completed.';
    MSG_TXOK            = '226 Transfer completed.';
    MSG_RXOK			= '226 Transfer completed.';
    MSG_ABORT           = '226 Operation aborted.';
    MSG_LOGGED          = '230 User %s logged in.';
	MSG_DELEOK			= '250 File %s deleted.';
	MSG_RMDOK 			= '250 Directory %s removed.';
    MSG_RNTOOK          = '250 File %s renamed to %s.';
	MSG_PWDOK			= '257 Current directory is "%s" (%d Mb free).';

    // 300
    MSG_NEEDPASS        = '331 User %s accepted. Password needed.';
    MSG_RNFROK          = '350 File %s exists. Issue RNTO next.';

    // 400
    MSG_LOWMEM          = '421 Insufficient resources. Please try again later.';
    MSG_CANTCONNECT     = '425 Cannot open data connection.';
    MSG_CONNBROKEN      = '446 Connection closed. Transfer aborted.';

    // 500
	MSG_BADCMD    		= '500 Unknown command "%s".';
    MSG_BADTYPE         = '501 Unknown data type "%s".';
    MSG_BADPORT         = '501 Bad PORT command. Usage: PORT n,n,n,n, n,n.';
    MSG_NOTIMPLEMENTED  = '502 Command "%s" not implemented yet.';
    MSG_USERFIRST       = '503 You are not logged in. Use USER command first.';
    MSG_NOTLOGGED       = '530 You are not logged in. Use USER and PASS commands first.';
    MSG_ALREADYLOGGED   = '550 Already Logged-in. Disconnect first.';
	MSG_BADLOGIN  		= '550 User %s cannot log in.';
    MSG_NOPERM    		= '550 Access denied.';
    MSG_CANTOPEN        = '550 Cannot open %s.';
    MSG_BADDELE         = '550 Cannot delete file %s.';
    MSG_BADRNTO         = '550 Cannot rename %s to %s.';
    MSG_WRITEERR        = '552 Error writing file %s. Transfer aborted.';
    MSG_NODIR           = '553 Path %s does not exist or cannot be accessed.';
    MSG_CANTMAKE        = '553 Cannot create %s.';

Var
	ServerList: PCollection;

{ TFTPConnection }

constructor TFTPConnection.Create(Socket: Integer;
  RemotePeer: TInetSockAddr; RootPath: AnsiString);
begin
	inherited Create;
    FFlags := [];
    FLastComm := '';
    FCtrlSocket := Socket;
    FDataSocket := 0;
    FRemotePeer := RemotePeer;
    FCloseTimeout := GetMSCounter + CLOSE_TIMEOUT;
    FCurrentFileHandle := 0;
    FUserName := '';
    FillChar(FDataPort, SizeOf(FDataPort), 0);
    FRootPath := GoodDir(RootPath);
    FCurrentPath := GoodDir(FRootPath);
    FCurrentFileName := '';
    FStatus := stBanner;
end;

destructor TFTPConnection.Destroy;
begin
    if (FStatus <> stDone) then SendResponse(MSG_BYE);
    if FDataSocket <> 0 then CloseSocket(FDataSocket);
    if FCtrlSocket <> 0 then CloseSocket(FCtrlSocket);
    if FCurrentFileHandle <> 0 then FileClose(FCurrentFileHandle);
	inherited Destroy;
end;

procedure TFTPConnection.Work;
begin
    case FStatus of
        stDone: Exit;
        stBanner: SendBanner;
        stWaitingUser: ;
        stWaitingPassword: ;
        stWaitingCommand: ;
        stSendingDir: SendDirBlock;
        stSendingList: SendListBlock;
        stReceivingFile: ReceiveFileBlock;
        stSendingFile: SendFileBlock;
    end;
    // Check for inactivity Timeout, and test connections
    if GetMSCounter > FCloseTimeout then FStatus := stDone
      else CheckControlSocket;
end;

procedure TFTPConnection.SendBanner;
begin
    SendResponse(MSG_BANNER);
    FStatus := stWaitingUser;
end;

procedure TFTPConnection.SendDirBlock;
Var
    DT: DateTime;
	Line: AnsiString;
begin
    Activity;
    if (FSearchRec.Name <> '.') and (FSearchRec.Name <> '..') then begin
        // Send Directory Entry
        UnpackTime(FSearchRec.Time, DT);
        if (FSearchRec.Attr and faDirectory) = faDirectory then
          Line := 'd' else Line := '-';
        if FSearchRec.Attr and faReadOnly = faReadOnly then
          Line := Line + 'r--r--r--' else Line := Line + 'rw-rw-rw-';
        Line := Line +
          '   1 owner    group    ' +
          PadL(IntToStr(FSearchRec.Size), 12, ' ') + ' ' +
          ShortMonthNames[DT.Month] + ' ' +
          PadL(IntToStr(DT.Day), 2, ' ') + ' ' +
          PadL(IntToStr(DT.Hour), 2, ' ') + ':' +
          PadL(IntToStr(DT.Min), 2, '0') + ' ' +
          LowerCase(FSearchRec.Name) +
          CRLF;
        Send(FDataSocket, Line[1], Length(Line), 0);
    end;
    // Search for Next Entry
    if FindNext(FSearchRec) <> 0 then begin
	   	// No more entries to list
       	FindClose(FSearchRec);
        FStatus := stWaitingCommand;
        SendResponse(MSG_ENDDIR);
        if (FDataSocket <> FCtrlSocket) then begin
         	CloseSocket(FDataSocket);
            FDataSocket := 0;
        end;
    end;
end;

procedure TFTPConnection.SendListBlock;
Var
	Line: AnsiString;
begin
	Activity;
    if (FSearchRec.Name <> '.') then begin
    	// Send List Entry.
        Line :=
          LowerCase(FSearchRec.Name) +
          CRLF;
          Send(FDataSocket, Line[1], Length(Line), 0);
    end;
    // Search for Next entry
	if FindNext(FSearchRec) <> 0 then begin
     	// No more entries to list
      	FindClose(FSearchRec);
        FStatus := stWaitingCommand;
        SendResponse(MSG_ENDDIR);
        if (FDataSocket <> FCtrlSocket) then begin
          	CloseSocket(FDataSocket);
            FDataSocket := 0;
        end;
    end;
end;

procedure TFTPConnection.SendFileBlock;
Var
    Len: LongInt;
    Close: Boolean;
    Buffer: Array[0..2047] of char;
begin
	Activity;
    Close := False;
    Len := FileRead(FCurrentFileHandle, buffer, SizeOf(Buffer));
    if Send(FDataSocket, Buffer, Len, 0) < len then begin
        SendResponse(MSG_CONNBROKEN);
        Close := True;
    end else begin
    	if Len < SizeOf(Buffer) then begin
            SendResponse(MSG_TXOK);
            Close := True;
        end;
    end;
    if Close then begin
        FileClose(FCurrentFileHandle);
        FCurrentFileHandle := 0;
        FStatus := stWaitingCommand;
        if FDataSocket <> FCtrlSocket then begin
        	CloseSocket(FDataSocket);
            FDataSocket := 0;
        end;
    end;
end;

procedure TFTPConnection.ReceiveFileBlock;
Var
	Len: LongInt;   	
    Close: Boolean;
    Buffer: Array[0..2047] of char;
begin
	Close := False;
	ioctlsocket(FDataSocket, FIONREAD, @Len);
    if Len > 0 then begin
    	Activity;
    	if Len > SizeOf(Buffer) then Len := SizeOf(Buffer);
        Len := Recv(FDataSocket, Buffer, Len, 0);
        if FileWrite(FCurrentFileHandle, buffer, len) < Len then begin
            SendResponse(Format(MSG_WRITEERR, [UnixFileName(FCurrentFileName)]));
            FileClose(FCurrentFileHandle);
            DeleteFile(FCurrentFileName);
            Close := True;
        end else if recv(FDataSocket, buffer, 0, 0) < 0 then begin
        	SendResponse(MSG_RXOK);
            FileClose(FCurrentFileHandle);
            Close := True;
        end;
    end;
    if Close then begin
        FCurrentFileHandle := 0;
        FCurrentFileName := '';
        if FDataSocket <> FCtrlSocket then begin
            CloseSocket(FDataSocket);
            FDataSocket := 0;
        end;
        FStatus := stWaitingCommand;
    end;
end;

procedure TFTPConnection.CheckControlSocket;
Var
    Len: LongInt;
    Line: AnsiString;
begin
    // Still Connected?
    if Recv(FCtrlSocket, Len, 0, 0) < 0 then begin
        FStatus := stDone;
        Exit;
    end;
    // Any commands to read?
    ioctlsocket(FCtrlSocket, FIONREAD, @len);
    if Len > 0 then begin
        Activity;
    	SetLength(Line, Len);
        Len := recv(FCtrlSocket, Line[1], Len, 0);
        SetLength(Line, Len);
        ProcessControlLine(Line);
    end;
end;

procedure TFTPConnection.SendResponse(Response: AnsiString);
begin
    if FCtrlSocket = 0 then Exit;
    Response := Response + CRLF;
	WriteLn(Response);
    Send(FCtrlSocket, Response[1], Length(Response), 0);
end;

procedure TFTPConnection.ProcessControlLine(Line: AnsiString);
Var
    i: Integer;
    Cmd, Params: AnsiString;
begin
	Write(Line);
    // Separate Command and Params
    i := Pos(' ', Line);
    if i = 0 then begin
    	Cmd := UpperCase(Trim(Line));
        Params := '';
    end else begin
    	Cmd := UpperCase(Trim(Copy(Line, 1, i - 1)));
        Params := Trim(Copy(Line, i + 1, Length(Line)));
    end;

    // Check Command
    if      Cmd = 'SYST' then SendResponse(MSG_SYST)
    else if Cmd = 'NOOP' then SendResponse(MSG_NOOP)
    else if Cmd = 'HELP' then Proc_Help(Params)
    else if Cmd = 'USER' then Proc_User(Params)
    else if Cmd = 'PASS' then Proc_Pass(Params)
    else if Cmd = 'PORT' then Proc_Port(Params)
    else if Cmd = 'TYPE' then Proc_Type(Params)
    else if (Cmd = 'PWD') or (Cmd = 'XPWD') then Proc_Pwd(Params)
    else if Cmd = 'CDUP' then Proc_Cwd('..')
    else if (Cmd = 'CWD') or (Cmd = 'XCWD') then Proc_Cwd(Params)
    else if (Cmd = 'MKD') or (Cmd = 'XMKD') then Proc_Mkd(Params)
    else if (Cmd = 'RMD') or (Cmd = 'XRMD') then Proc_Rmd(Params)
    else if Cmd = 'RNFR' then Proc_RnFr(Params)
    else if Cmd = 'RNTO' then Proc_Rnto(Params)
    else if Cmd = 'LIST' then Proc_List(Params)
    else if Cmd = 'NLST' then Proc_Nlst(Params)
    else if Cmd = 'STOR' then Proc_Stor(Params)
    else if Cmd = 'DELE' then Proc_Dele(Params)
    else if Cmd = 'RETR' then Proc_Retr(Params)
    else if Cmd = 'SITE' then Proc_Site(Params)
    else if Cmd = 'SIZE' then Proc_Size(Params)
    else if Cmd = 'QUIT' then Proc_Quit(Params)
    else if Cmd = 'ABOR' then Proc_Abor(Params)
    else SendResponse(Format(MSG_NOTIMPLEMENTED, [Cmd]));
end;

procedure TFTPConnection.Activity;
begin
    FCloseTimeout := GetMSCounter + CLOSE_TIMEOUT;
end;

procedure TFTPConnection.AdjustPath(var Path: AnsiString);
var
	i: Integer;
begin
    While (Length(Path) > 0) and (Path[Length(Path)] in ['\', '/']) do
      SetLength(Path, Length(Path) - 1);
	for i := 1 to Length(Path) do if Path[i] = '/' then Path[i] := '\';
    if Copy(Path, 1, 1) = '\' then
        Path := GoodDir(FRootPath) + Copy(Path, 2, Length(Path))
    else begin
        Path := GoodDir(FCurrentPath) + Path;
    end;
    Path := LowerCase(Path);
end;

function TFTPConnection.UnixFileName(Path: AnsiString): AnsiString;
Var
  	i : integer;
begin
	Path := LowerCase(Path);
	if copy(Path, 2, 1) = ':' then Delete(Path, 1, 2);
	for i := 1 to Length(Path) do begin
        if Path[i]='\' then Path[i] := '/';
        if (Length(Path) > 1) and (Copy(Path, Length(Path), 1) = '/') then
          Delete(Path, Length(Path), 1);
    end;
    Result := Path;
end;

function TFTPConnection.OpenDataConnection: Boolean;
begin
    if FDataSocket <> 0 then begin
        // Already connected
        SendResponse(MSG_ALREADYCONNECT);
        Result := True;
    end else if FDataPort.Addr = 0 then begin
    	// No PORT issued
        SendResponse(MSG_ALREADYCONNECT);
        FDataSocket := FCtrlSocket;
	    Result := True;
    end else begin
    	// Port ISSUED. Connect.
        SendResponse(MSG_CONNECTING);
        FDataPort.Family := AF_INET;
        FDataSocket := Socket(AF_INET, SOCK_STREAM, 0);
        Result := Connect(FDataSocket, FDataPort, SizeOf(FDataPort));
        if not Result then begin
            SendResponse(MSG_CANTCONNECT);
            CloseSocket(FDataSocket);
            FDataSocket := 0;
        end;
    end;
end;

Procedure TFTPConnection.Proc_Pwd(Params: AnsiString);
Var
	Drive: Integer;
	FreeSpace: LongInt;
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    if Pos(':', FCurrentPath) > 0 then
    	    Drive := Ord(UpCase(FCurrentPath[1])) - 64
	    else begin
    		Drive := 0;
	    end;
    	FreeSpace := DiskFree(Drive) div (1024 * 1024);
        SendResponse(Format(MSG_PWDOK, [UnixFilename(FCurrentPath), FreeSpace]));
    end;
end;

Procedure TFTPConnection.Proc_Cwd(Params: AnsiString);
var
    i : byte;
    Rec: TSearchRec;
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	if IsDir(Params) then begin
	    	FCurrentPath := Params;
    	    Proc_Pwd('');
	    end else begin
    	    SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Mkd(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	if CreateDir(Params) then
	    	SendResponse(Format(MSG_MKDOK, [UnixFileName(Params)]))
    	else begin
	    	SendResponse(Format(MSG_CANTMAKE, [UnixFileName(Params)]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Rmd(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	if RemoveDir(Params) then
	    	SendResponse(Format(MSG_RMDOK, [UnixFileName(Params)]))
    	else begin
	    	SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Help(Params: AnsiString);
begin
    SendResponse(
      '214-Valid commands are:' + CRLF +
      '214-' + CRLF +
      '214-   USER    PASS    PORT    TYPE    PWD    XPWD    CDUP    CWD' + CRLF +
      '214-   XCWD    LIST    HELP    SYST    STOR   DELE    NLST    MKD' + CRLF +
      '214    XMKD    RMD     XRMD    RNFR    RNTO');
end;

Procedure TFTPConnection.Proc_User(Params: AnsiString);
begin
    if FStatus = stWaitingUser then begin
    	FUserName := UpperCase(Params);
        SendResponse(Format(MSG_NEEDPASS, [FUserName]));
        FStatus := stWaitingPassword;
    end else begin
        SendResponse(MSG_ALREADYLOGGED);
    end;
end;

Procedure TFTPConnection.Proc_Pass(Params: AnsiString);
begin
    if FStatus < stWaitingPassword then
        SendResponse(MSG_USERFIRST)
    else if FStatus > stWaitingPassword then
        SendResponse(MSG_ALREADYLOGGED)
    else begin
	    SendResponse(Format(MSG_LOGGED, [FUserName]));
        FStatus := stWaitingCommand;
    end;
//	SendResponse(Format(MSG_BADLOGIN, [UserName]));
//  FStatus := stWaitingUser;
end;

Procedure TFTPConnection.Proc_Port(Params: AnsiString);
begin
    // The incomming string is something like "AAA,BBB,CCC,DDD,EEE,FFF"
    // where AAA..DDD is the IP Address of the remote client, and
    // EEE and FFF represent the High and Low bytes of the client's port.
    FDataPort.Addr := IP_StringToAddr(
      ParseParamByNumber(Params, 1, ',') + '.' +
      ParseParamByNumber(Params, 2, ',') + '.' +
      ParseParamByNumber(Params, 3, ',') + '.' +
      ParseParamByNumber(Params, 4, ',')
      );
    FDataPort.port := Swap(Word(
      IVal(ParseParamByNumber(Params, 5, ',')) * $100 +
      IVal(ParseParamByNumber(Params, 6, ','))
      ));
    if (FDataPort.addr <> 0) and (FDataPort.port <> 0) then begin
        if FDataSocket <> 0 then CloseSocket(FDataSocket);
        FDataSocket := 0;
        SendResponse(MSG_PORTOK);
    end else begin
        SendResponse(MSG_BADPORT);
    end;
end;

Procedure TFTPConnection.Proc_Type(Params: AnsiString);
begin
	Params := Uppercase(Params);
    if (Params = 'BINARY') or (Params = 'IMAGE') or (Params = 'I') then
        Include(FFlags, ftpBinary)
    else if (Params = 'ASCII') or (Params = 'A') or (Params = 'TEXT') then
        Exclude(FFlags, ftpBinary)
    else begin
    	SendResponse(Format(MSG_BADTYPE, [Params]));
        Exit
    end;
    SendResponse(Format(MSG_TYPEOK, [Params]));
end;

Procedure TFTPConnection.Proc_RnFr(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	if FileExists(Params) then begin
	        SendResponse(MSG_RNFROK);
    	    FCurrentFileName := Params;
	    end else begin
    		SendResponse(Format(MSG_CANTOPEN, [UnixFileName(Params)]));
	        FCurrentFileName := '';
        end;
    end;
end;

Procedure TFTPConnection.Proc_Rnto(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	if FileExists(FCurrentFileName)
	      and RenameFile(FCurrentFileName, Params) then
    	    SendResponse(Format(MSG_RNTOOK,
	          [UnixFileName(FCurrentFileName), UnixFileName(Params)]))
    	else begin
	        SendResponse(Format(MSG_BADRNTO,
    	      [UnixFileName(FCurrentFileName), UnixFileName(Params)]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_List(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    if Params = '' then Params := '*.*';
    	AdjustPath(Params);
	    if FindFirst(Params, AnyFile - VolumeID - Hidden, FSearchRec) = 0 then begin
            if OpenDataConnection then FStatus := stSendingDir;
	    end else begin
    		FindClose(FSearchRec);
	        SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Nlst(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    if Params = '' then Params := '*.*';
    	AdjustPath(Params);
	    if FindFirst(Params, AnyFile - VolumeID - Hidden, FSearchRec) = 0 then begin
    	    if OpenDataConnection then FStatus := stSendingList;
	    end else begin
    		FindClose(FSearchRec);
	        SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Stor(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	FCurrentFileHandle := FileCreate(Params);
	    if FCurrentFileHandle < 0 then begin
    		SendResponse(Format(MSG_CANTMAKE, [UnixFileName(Params)]));
	    end else begin
    	    if OpenDataConnection then begin
	        	FStatus := stReceivingFile;
    	        FCurrentFileName := Params;
	        end else begin
    	        FileClose(FCurrentFileHandle);
	            FCurrentFileHandle := 0;
    	    end;
        end;
    end;
end;

Procedure TFTPConnection.Proc_Dele(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	if DeleteFile(Params) then
	        SendResponse(Format(MSG_DELEOK, [UnixFileName(Params)]))
    	else begin
	        SendResponse(Format(MSG_BADDELE, [UnixFileName(Params)]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Retr(Params: AnsiString);
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    AdjustPath(Params);
    	FCurrentFileHandle := FileOpen(Params, fmOpenRead);
	    if FCurrentFileHandle < 0 then
    	    SendResponse(Format(MSG_CANTOPEN, [UnixFileName(Params)]))
	    else begin
		    if OpenDataConnection then begin
	            FStatus := stSendingFile;
		        FCurrentFileName := Params;
	    	end else begin
	            FileClose(FCurrentFileHandle);
    	        FCurrentFileHandle := 0;
        	end;
        end;
    end;
end;

Procedure TFTPConnection.Proc_Site(Params: AnsiString);
Var
	Cmd: AnsiString;
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
	    Cmd := UpperCase(ParseParamByNumber(Params, 1, ' '));
    	if Cmd = 'HELP' then
	    	Proc_Help(Trim(Copy(Params, 5, Length(Params))))
    	else begin
	        SendResponse(Format(MSG_NOTIMPLEMENTED, ['SITE ' + Params]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Size(Params: AnsiString);
Var
	h: Integer;
    nSize: LongInt;
begin
    if FStatus < stWaitingCommand then begin
        SendResponse(MSG_NOTLOGGED);
    end else begin
    	AdjustPath(Params);
        h := FileOpen(Params, fmOpenRead);
        if h < 0 then begin
            SendResponse(Format(MSG_CANTOPEN, [UnixFileName(Params)]));
        end else begin
            nSize := FileSeek(h, 0, fsFromEnd);
            FileClose(h);
            SendResponse(Format(MSG_SIZE, [nSize]));
        end;
    end;
end;

Procedure TFTPConnection.Proc_Quit(Params: AnsiString);
begin
    FStatus := stDone;
end;

Procedure TFTPConnection.Proc_Abor(Params: AnsiString);
begin
	// Close Data Socket
    SendResponse(MSG_ABORT);
    if (FDataSocket <> 0) then begin
        if (FDataSocket <> FCtrlSocket) then CloseSocket(FDataSocket);
	    FDataSocket := 0;
    end;
    // Close any open File Handle
    if FCurrentFileHandle <> 0 then FileClose(FCurrentFileHandle);
    FCurrentFileHandle := 0;
    // Delete any incomming file
    if FStatus = stReceivingFile then DeleteFile(FCurrentFileName);
    // Done
    FStatus := stWaitingCommand;
end;


{ TFTPServer }
constructor TFTPServer.Create;
begin
    inherited Create;
    New(FConnections, Init(10, 1));
    FSocket := 0;
    GetDir(0, FRootPath);
    ServerList.Insert(Self);
end;

destructor TFTPServer.Destroy;
Var
    Conn: TFTPConnection;
begin
    ServerList.Delete(Self);
    if FSocket > 0 then CloseSocket(FSocket);
    While FConnections.Count > 0 do begin
        Conn := TFTPConnection(FConnections.At(0));
        Conn.Free;
        FConnections.AtDelete(0);
    end;
    Dispose(FConnections, Done);
end;

Function TFTPServer.Start(Address: TIPAddress; Port: WORD; RootPath: AnsiString): Boolean;
Var
    Addr: TInetSockAddr;
begin
	Result := False;
    if FSocket > 0 then Exit;
    // Open Server Socket to accept incoming connections
	Addr.Family   := PF_INET;
    Addr.Port     := Port;
    Addr.Addr     := Address;
    FSocket 	  := Socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
    if (FSocket <= 0) or not Bind(FSocket, Addr, SizeOf(Addr))
      or not Listen(FSocket, 10) then begin
        if FSocket <> 0 then CloseSocket(FSocket);
        FSocket := 0;
        Exit;
    end;
    // Done
    FRootPath := RootPath;
    Result := True;
end;

procedure TFTPServer.AddConnection(Socket: Integer; Addr: TInetSockAddr);
Var
	Conn: TFTPConnection;
begin
    Conn := TFTPConnection.Create(Socket, Addr, FRootPath);
    FConnections.Insert(Conn);
end;

procedure TFTPServer.Work;
Var
    i: Integer;
    HSock, L: LongInt;
    Addr: TInetSockAddr;
    Conn: TFTPConnection;
begin
    // Look for a new incomming connection
    L := SizeOf(Addr);
    HSock := Accept(FSocket, Addr, L);
    if HSock > 0 then AddConnection(HSock, Addr);
    // Make active connections work
    i := 0;
    While i < FConnections.Count do begin
        Conn := FConnections.At(i);
        Conn.Work;
        // If the connection is done, free it
        if Conn.Status = stDone then begin
            Conn.Free;
            FConnections.AtDelete(i);
        end else begin
            Inc(i);
        end;
    end;
end;

procedure FTP_Work;
Var
	i: Integer;
begin
    NET_Work_Low;
    for i := 0 to ServerList.Count - 1 do begin
        TFTPServer(ServerList.At(i)).Work;
    end;
end;


Var
    Server, i: Integer;
initialization
    // Creamos la lista de Servers
    New(ServerList, Init(10, 1));
    // Agregamos un procedure de Idle
	NET_AddIdleProc(@FTP_Work);
finalization
	// Quitamos el procedure de Idle
    NET_RemoveIdleProc(@FTP_Work);
    // Quitamos los Servers
    for i := 0 to ServerList.Count - 1 do TFTPServer(ServerList.At(i)).Free;
    // Eliminamos la lista de servers
    Dispose(ServerList, Done);
end.