Unit UDP;

interface

Uses
    Network, Sockets, DosSock, IP, Objects;

Function UDP_Init: Boolean;
Function UDP_Deinit: Boolean;
{ Manejo de Sockets UDP }
Function UDP_Socket				(Domain, SocketType, Protocol: Longint): Pointer;
Function UDP_Bind 				(Sock: Pointer; Var Addr; AddrLen: Longint): Boolean;
Function UDP_Connect    		(Sock: Pointer; Var Addr; Addrlen: Longint): Boolean;
Function UDP_Send 				(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
Function UDP_Recv 				(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
Function UDP_Close      		(Sock: Pointer): LongInt;
Function UDP_GetLastError: LongInt;
Function UDP_ioctlsocket		(Sock: Pointer; Option: LongInt; d: Pointer): LongInt;
Function UDP_Listen				(Sock: Pointer; MaxConnect: Longint): Boolean;
Function UDP_Accept     		(Sock: Pointer; Var Addr; Var Addrlen: Longint): Pointer;
Function UDP_GetSocketName  	(Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Function UDP_GetPeerName    	(Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Function UDP_Shutdown       	(Sock: Pointer; How: Longint): Longint;
Function UDP_GetSocketOptions	(Sock: Pointer; Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;
Function UDP_SetSocketOptions	(Sock: Pointer; Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;

implementation

Type
	PUDPHeader = ^TUDPHeader;
	TUDPHeader = packed record
        SourcePort: WORD;
        TargetPort: WORD;
        MessageLength: WORD;
        Checksum: WORD;
    end;

    TUDPBuff = record
        RemoteIP: TIPAddress;
        RemotePort: Integer;
        BufferData: AnsiString;
    end;

	PSocketRec = ^TSocketRec;
    TSocketRec = record
        LocalIP: TIPAddress;
        LocalPort: WORD;
        RemoteIP: TIPAddress;
        RemotePort: WORD;
        BufferCount: Integer;
        Buffers: Array[1..10] of TUDPBuff;
    end;

Var
	LastError: LongInt;
    SockList: PCollection;

procedure ProcessIP(SourceIP, TargetIP: TIPAddress;
  Var Data; DataLen: Word);
Var
    i: Integer;
    Hdr: PUDPHeader;
    PSock: PSocketRec;
begin
    Hdr := @Data;
    if Hdr^.TargetPort = 0 then Exit;
    { Lleg un datagrama UDP. Si algn socket lo quiere, lo colocamos }
    { en el buffer del socket correspondiente. Si no, lo descartamos. }
    for i := 0 to SockList^.Count - 1 do begin
    	PSock := PSocketRec(SockList^.At(i));
        if (PSock^.LocalPort = Hdr^.TargetPort) then begin
            { Este es el socket que atiende este puerto local. }
            if PSock^.BufferCount < High(PSock^.Buffers) then begin
                Inc(PSock^.BufferCount);
                With PSock^.Buffers[PSock^.BufferCount] do begin
                    RemoteIP := SourceIP;
                    RemotePort := Hdr^.SourcePort;
                    Dec(DataLen, SizeOf(TUDPHeader));
                    SetLength(BufferData, DataLen);
                    Move(PChar(@Data)[SizeOf(TUDPHeader)], BufferData[1], DataLen);
                end;
            end;
        	break;
        end;
    end;
    { Debug }
(*	WriteLn('  *UDP*');
	WriteLn('   Source Port   : ', Swap(Hdr^.SourcePort));
	WriteLn('   Target Port   : ', Swap(Hdr^.TargetPort));
   	WriteLn('   Message Length: ', Swap(Hdr^.MessageLength));
	WriteLn('   Checksum      : ', IntToHex(Hdr^.Checksum);
    Write  ('   Data          : ');
    { Tomamos los Datos }
   	Dec(DataLen, SizeOf(TUDPHeader));
	PData := @Data;
	Inc(PData, SizeOf(TUDPHeader));
    for i := 0 to DataLen - 1 do Write(PData[i]);
	WriteLn('(fin)');	*)
end;

Function GetFreePort: Integer;
Var
	Resu: WORD;
    i, Prt: Integer;
begin
	for Prt := 1024 to 8192 do begin
    	Resu := Prt;
        for i := 0 to SockList^.Count - 1 do begin
        	if PSocketRec(SockList^.At(i))^.LocalPort = Prt then begin
            	Resu := 0;
                break;
			end;
        end;
        if Resu <> 0 then begin
        	GetFreePort := Swap(Resu);
            Exit;
        end;
    end;
	GetFreePort := 0;
end;

Function UDP_Init: Boolean;
Var
	PS: TProtocolStruct;
begin
	Result := IP_AddProcessor(IPPROTO_UDP, @ProcessIP);
    {}
    PS.Domain     			:= PF_INET;
    PS.SocketType           := SOCK_DGRAM;
    PS.Protocol   			:= IPPROTO_IP;
    PS.FNSocket             := @UDP_Socket;
    PS.FNClose              := @UDP_Close;
    PS.FNGetLastError   	:= @UDP_GetLastError;
    PS.FNSend           	:= @UDP_Send;
    PS.FNRecv       	    := @UDP_Recv;
    PS.FNBind   	        := @UDP_Bind;
    PS.FNListen             := @UDP_Listen;
    PS.FNAccept             := @UDP_Accept;
    PS.FNConnect    	    := @UDP_Connect;
    PS.FNShutdown           := @UDP_Shutdown;
    PS.FNGetSocketName      := @UDP_GetSocketName;
    PS.FNGetPeerName        := @UDP_GetPeerName;
    PS.FNSetSocketOptions   := @UDP_SetSocketOptions;
    PS.FNGetSocketOptions   := @UDP_GetSocketOptions;
    PS.FNioctlsocket        := @UDP_ioctlsocket;
    Result := Result and AddProtocol(PS);
    PS.Protocol             := IPPROTO_UDP;
    Result := Result and AddProtocol(PS);
end;

Function UDP_GetLastError: LongInt;
begin
    UDP_GetLastError := LastError;
end;

Function UDP_Deinit: Boolean;
begin
	UDP_Deinit := IP_RemoveProcessor(IPPROTO_UDP);
end;

Function UDP_Socket(Domain, SocketType, Protocol: Longint): Pointer;
Var
	PSock: PSocketRec;
begin
    New(PSock);
    PSock^.LocalIP     := 0;
    PSock^.LocalPort   := 0;
    PSock^.RemoteIP    := 0;
    PSock^.RemotePort  := 0;
    PSock^.BufferCount := 0;
    UDP_Socket         := PSock;
    SockList^.Insert(PSock);
end;

Function UDP_Bind(Sock: Pointer; Var Addr; AddrLen: Longint): Boolean;
Var
    PSock: PSocketRec;
    InAddr: TInetSockAddr;
begin
    UDP_Bind := False;
    PSock := PSocketRec(Sock);
    InAddr := TInetSockAddr(Addr);
    { Check the given Address }
    if InAddr.family <> AF_INET then begin
        LastError := WSAEFAULT;
        Exit;
    end;
	{ If it's necessary, we choose a free local port for the socket }
    if InAddr.port = 0 then InAddr.port := GetFreePort;
    { Bind the socket to the local IP and Address }
    PSock^.LocalIP := InAddr.addr;
    PSock^.LocalPort := InAddr.port;
    { Done }
	UDP_Bind := True;
end;

Function UDP_Connect(Sock: Pointer; Var Addr; Addrlen: Longint): Boolean;
Var
    PSock: PSocketRec;
    InAddr: TInetSockAddr;
begin
	PSock := PSocketRec(Sock);
    InAddr := TInetSockAddr(Addr);
    { Check the given Address }
    if InAddr.family <> AF_INET then begin
    	UDP_Connect := False;
        LastError := WSAEFAULT;
        Exit;
    end;
    { If the socket is not bound to a local IP, now it's the time to do it }
(*    if (PSock^.LocalIP = INADDR_ANY) then begin
        if InAddr.Addr = INADDR_BROADCAST then begin
        end else begin
        end;
        PSock^.LocalIP := IP_GetLocalAddressForTarget(InAddr.addr);
        if PSock^.LocalIP = 0 then begin
            UDP_Connect := False;
            LastError := WSAENETUNREACH;
        	Exit;
        end;
    end;

    or (PSock^.LocalIP = 0) then begin
    	
    end;
*)
    { If the socket is not bound to a local port, now it's the time to do it }
    if PSock^.LocalPort = 0 then PSock^.LocalPort := GetFreePort;
    { Bind the socket to the remote peer }
    PSock^.RemoteIP := InAddr.addr;
    PSock^.RemotePort := InAddr.port;
    { Done }
    UDP_Connect := True;
end;

Function UDP_Send(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
Var
	Resp: LongInt;
	P: PUDPHeader;
    PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
	{ Create de UDP datagram }
    GetMem(P, AddrLen + SizeOf(TUDPHeader));
    P^.SourcePort := PSock^.LocalPort;
    P^.TargetPort := PSock^.RemotePort;
    P^.MessageLength := Swap(Word(AddrLen + SizeOf(TUDPHeader)));
    P^.Checksum   := 0;
    Move(Addr, PChar(P)[SizeOf(TUDPHeader)], AddrLen);
    { Send the UDP datagram }
    Resp := IP_Send(PSock^.RemoteIP, IPPROTO_UDP, P^, AddrLen + SizeOf(TUDPHeader));
	if Resp = 0 then begin
        UDP_Send := AddrLen;
    end else begin
        UDP_Send  := SOCKET_ERROR;
        LastError := Resp;
    end;
    FreeMem(P, AddrLen + SizeOf(TUDPHeader));
end;

Function UDP_Recv(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
Var
    PSock: PSocketRec;
    i, BuffSize: Integer;
begin
    PSock := PSocketRec(Sock);
    { If there's nothing to be read, we return zero }
    if (PSock^.BufferCount = 0) then begin
        UDP_Recv := 0;
        LastError := 0;
        Exit;
    end;
    { If the supplied buffer is not large enough, we return SOCKET_ERROR, }
    { setting LastError to WSAEMSGSIZE.                                   }
    BuffSize := Length(PSock^.Buffers[1].BufferData);
    if AddrLen < BuffSize then begin
        UDP_Recv  := SOCKET_ERROR;
        LastError := WSAEMSGSIZE;
        Exit;
    end;
    { Read data from the first available datagram, and copy it to the     }
    { user-supplied buffer.                                               }
    Move(PSock^.Buffers[1].BufferData[1], Addr, BuffSize);
    { If necesary, we remove the datagram from the buffer queue           }
    if (Flags and MSG_PEEK) <> MSG_PEEK then begin
	    PSock^.Buffers[1].BufferData := '';
    	Dec(PSock^.BufferCount);
	    for i := 1 to PSock^.BufferCount do begin
    	    PSock^.Buffers[i] := PSock^.Buffers[i + 1];
        end;
    end;
    { Done }
    UDP_Recv := BuffSize;
    LastError := 0;
end;

Function UDP_Close(Sock: Pointer): LongInt;
Var
    PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
    SockList^.Delete(PSock);
    Dispose(PSock);
    UDP_Close := 0;
    LastError := 0;
end;

Function UDP_ioctlsocket(Sock: Pointer; Option: LongInt; d: Pointer): LongInt;
Var
	dw: DWORD;
    PSock: PSocketRec;
begin
    NET_Work_Low;
    UDP_ioctlsocket := 0;
    PSock := PSocketRec(Sock);
    case Option of
        FIONREAD:
        	begin
                if (PSock^.BufferCount > 0) then
                    dw := Length(PSock^.Buffers[1].BufferData)
                else begin
                	dw := 0;
                end;
                Move(dw, d^, 4);
            end;
        else
            begin
                UDP_ioctlsocket := SOCKET_ERROR;
                LastError := WSAEINVAL;
            end;
    end;
end;

Function UDP_Listen(Sock: Pointer; MaxConnect: Longint): Boolean;
begin
    UDP_Listen := False;
    LastError := WSAEOPNOTSUPP;
end;

Function UDP_Accept(Sock: Pointer; Var Addr; Var Addrlen: Longint): Pointer;
begin
    UDP_Accept := nil;
    LastError := WSAEOPNOTSUPP;
end;

Function UDP_GetSocketName(Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Var
    PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
    { Check to see if there's enough room in the destination buffer }
    if AddrLen < SizeOf(TInetSockAddr) then begin
        UDP_GetSocketName := SOCKET_ERROR;
        LastError := WSAEMSGSIZE;
    	Exit;
    end;
    { Copy the local IP and Port to the destination buffer }
    TInetSockAddr(Addr).family := AF_INET;
    TInetSockAddr(Addr).addr   := PSock^.LocalIP;
    TInetSockAddr(Addr).port   := PSock^.LocalPort;
    AddrLen := SizeOf(TInetSockAddr);
    { Done }
    UDP_GetSocketName := 0;
end;

Function UDP_GetPeerName(Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Var
    PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
    { Check to see if there's enough room in the destination buffer }
    if AddrLen < SizeOf(TInetSockAddr) then begin
        UDP_GetPeerName := SOCKET_ERROR;
        LastError := WSAEMSGSIZE;
    	Exit;
    end;
    { Check to see if the socket is connected }
    if (PSock^.RemoteIP = 0) or (PSock^.RemotePort = 0) then begin
        UDP_GetPeerName := SOCKET_ERROR;
        LastError := WSAENOTCONN;
    	Exit;
    end;
    { Copy the remote IP and Port to the destination buffer }
    TInetSockAddr(Addr).family := AF_INET;
    TInetSockAddr(Addr).addr   := PSock^.RemoteIP;
    TInetSockAddr(Addr).port   := PSock^.RemotePort;
    AddrLen := SizeOf(TInetSockAddr);
    { Done }
    UDP_GetPeerName := 0;
end;

Function UDP_Shutdown(Sock: Pointer; How: Longint): Longint;
begin
    UDP_Shutdown := SOCKET_ERROR;
    LastError := WSAEINVAL;
end;

Function UDP_GetSocketOptions(Sock: Pointer; Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;
begin
    UDP_GetSocketOptions := SOCKET_ERROR;
    LastError := WSAEINVAL;
end;

Function UDP_SetSocketOptions(Sock: Pointer; Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;
begin
    UDP_SetSocketOptions := SOCKET_ERROR;
    LastError := WSAEINVAL;
end;




initialization
	LastError := 0;
    SockList  := New(PCollection, Init(10, 10));
finalization
    { If there's any unclosed socket, close it now }
    While SockList^.Count > 0 do UDP_Close(SockList^.At(0));
    { Free the socket collection }
    Dispose(SockList, Done);
end.
